from autocat.data.intermediates.nrr import NRR_MOLS, NRR_INTERMEDIATE_NAMES
from autocat.data.intermediates.orr import ORR_MOLS, ORR_INTERMEDIATE_NAMES

__all__ = ["NRR_MOLS", "NRR_INTERMEDIATE_NAMES", "ORR_MOLS", "ORR_INTERMEDIATE_NAMES"]



import os
from contextlib import contextmanager


@contextmanager
def change_working_dir(new_dir: str):
    current_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(current_dir)



import os
import numpy as np
from typing import List
from typing import Dict
from typing import Any
from typing import Sequence

from ase import Atoms
from ase.data import atomic_numbers
from ase.data import ground_state_magnetic_moments
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.adsorption import AdsorbateSiteFinder
from autocat.surface import generate_surface_structures


class AutocatSaaGenerationError(Exception):
    pass


def _find_dopant_index(structure, dopant_element):
    """Helper function for finding the index of the (single) dopant atom."""
    # TODO(@lancekavalsky): implement multi-dopant-atom indices
    #    # Find index of species with lowest count
    #    unique, counts = np.unique(syms, return_counts=True)
    #    ind = np.where(syms == unique[np.argmin(counts)])[0][0]
    symbols = np.array(structure.symbols)
    dopant_index = np.where(symbols == dopant_element)
    if np.size(dopant_index) < 1:
        msg = f"Dopant element {dopant_element} not found in structure"
        raise AutocatSaaGenerationError(msg)
    elif np.size(dopant_index) > 1:
        msg = f"More than one atom of {dopant_element} found in structure"
        raise NotImplementedError(msg)
    return dopant_index[0][0]


def generate_saa_structures(
    host_species: List[str],
    dopant_species: List[str],
    crystal_structures: Dict[str, str] = None,
    facets: Dict[str, str] = None,
    supercell_dim: Sequence[int] = (3, 3, 4),
    default_lat_param_lib: str = None,
    a_dict: Dict[str, float] = None,
    c_dict: Dict[str, float] = None,
    set_host_magnetic_moments: List[str] = None,
    host_magnetic_moments: Dict[str, float] = None,
    set_dopant_magnetic_moments: List[str] = None,
    dopant_magnetic_moments: Dict[str, float] = None,
    vacuum: float = 10.0,
    n_fixed_layers: int = 0,
    place_dopant_at_center: bool = True,
    write_to_disk: bool = False,
    write_location: str = ".",
    dirs_exist_ok: bool = False,
) -> Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]:
    """
    Builds single-atom alloys for all combinations of host species and dopant
    species given. Will write the structures to separate directories if
    specified.

    Parameters
    ----------

    host_species (REQUIRED):
        List of chemical species of desired host (substrate) species.

    dopant_species (REQUIRED):
        List of chemical symbols of desired single-atom dopant species.

    crystal_structures:
        Dictionary with crystal structure to be used for each species.
        These will be passed on as input to `ase.build.bulk`. So, must be one
        of sc, fcc, bcc, tetragonal, bct, hcp, rhombohedral, orthorhombic,
        diamond, zincblende, rocksalt, cesiumchloride, fluorite or wurtzite.
        If not specified, the default reference crystal structure for each
        species from `ase.data` will be used.

    facets:
        Dictionary with the surface facets to be considered for each
        species.
        If not specified for a given species, the following defaults will be
        used based on the crystal structure:
        fcc/bcc: 100, 111, 110
        hcp: 0001

    supercell_dim:
        Tuple or List specifying the size of the supercell to be
        generated in the format (nx, ny, nz).
        Defaults to (3, 3, 4).

    default_lat_param_lib:
        String indicating which library the lattice constants should be pulled
        from if not specified in either a_dict or c_dict.

        Options:
        pbe_fd: parameters calculated using xc=PBE and finite-difference
        beefvdw_fd: parameters calculated using xc=BEEF-vdW and finite-difference
        pbe_pw: parameters calculated using xc=PBE and a plane-wave basis set
        beefvdw_fd: parameters calculated using xc=BEEF-vdW and a plane-wave basis set

        N.B. if there is a species present in `host_species` that is NOT in the
        reference library specified, it will be pulled from `ase.data`.

    a_dict:
        Dictionary with lattice parameters <a> to be used for each species.
        If not specified, defaults from `default_lat_param_lib` are used.

    c_dict:
        Dictionary with lattice parameters <c> to be used for each species.
        If not specified, defaults from `default_lat_param_lib` are used.

    set_host_magnetic_moments:
        List of host species for which magnetic moments need to be set.
        If not specified, magnetic moments will be set only for Fe, Co, Ni
        (the ferromagnetic elements).

    host_magnetic_moments:
        Dictionary with the magnetic moments to be set for the host chemical
        species listed previously.
        If not specified, default ground state magnetic moments from
        `ase.data` are used.

    set_dopant_magnetic_moments:
        List of single-atom species for which magnetic moments need to be set.
        If not specified, magnetic moments will guessed for all dopant species from
        `ase.data`.

    dopant_magnetic_moments:
        Dictionary with the magnetic moments to be set for the single-atom
        dopant species listed previously.
        If not specified, default ground state magnetic moments from
        `ase.data` are used.

    vacuum:
        Float specifying the amount of vacuum (in Angstrom) to be added to
        the slab (the slab is placed at the center of the supercell).
        Defaults to 10.0 Angstrom.

    n_fixed_layers:
        Integer giving the number of layers of the slab to be fixed
        starting from the bottom up (e.g., a value of 2 will fix the
        bottom 2 layers).
        Defaults to 0 (i.e., no layers in the slab fixed).

    place_dopant_at_center:
        Boolean specifying whether the single-atom should be placed
        at the center of the unit cell. If False, the single-atom will
        be placed at the origin.
        Defaults to True.

    write_to_disk:
        Boolean specifying whether the bulk structures generated should be
        written to disk.
        Defaults to False.

    write_location:
        String with the location where the per-species/per-crystal structure
        directories must be constructed and structure files written to disk.
        In the specified write_location, the following directory structure
        will be created:
        [host]/[dopant]/[facet]/substrate/input.traj

    dirs_exist_ok:
        Boolean specifying whether existing directories/files should be
        overwritten or not. This is passed on to the `os.makedirs` builtin.
        Defaults to False (raises an error if directories corresponding the
        species and crystal structure already exist).

    Returns
    -------

    Dictionary with the single-atom alloy structures as `ase.Atoms` objects and
    write-location, if any, for each {crystal structure and facet} specified for
    each input host and dopant species combination.

    Example:
    {
        "Fe": {
            "Cu": {
                "bcc100": {
                    "structure": FeN-1_Cu1_saa_obj,
                    "traj_file_path": "/path/to/Cu/on/bcc/Fe/100/surface/traj/file"
                },
                "bcc110": ...,
            },
            "Ru": {
                ...
            },
        },
        "Rh": {
            ...
        }
    }

    """

    hosts = generate_surface_structures(
        host_species,
        crystal_structures=crystal_structures,
        facets=facets,
        supercell_dim=supercell_dim,
        default_lat_param_lib=default_lat_param_lib,
        a_dict=a_dict,
        c_dict=c_dict,
        set_magnetic_moments=set_host_magnetic_moments,
        magnetic_moments=host_magnetic_moments,
        vacuum=vacuum,
        n_fixed_layers=n_fixed_layers,
    )

    if set_dopant_magnetic_moments is None:
        set_dopant_magnetic_moments = dopant_species
    if dopant_magnetic_moments is None:
        dopant_magnetic_moments = {}

    dop_mm_library = {
        dop: ground_state_magnetic_moments[atomic_numbers[dop]]
        for dop in dopant_species
    }
    dop_mm_library.update(dopant_magnetic_moments)

    saa_structures = {}
    # iterate over hosts
    for host in hosts:
        saa_structures[host] = {}
        # iterate over single-atoms
        for dopant in dopant_species:
            # ensure host != single-atom
            if dopant == host:
                continue
            saa_structures[host][dopant] = {}
            # iterate over surface facets
            for facet in hosts[host]:
                host_structure = hosts[host][facet].get("structure")
                doped_structure = substitute_single_atom_on_surface(
                    host_structure,
                    dopant,
                    place_dopant_at_center=place_dopant_at_center,
                    dopant_magnetic_moment=dop_mm_library.get(dopant),
                )

                traj_file_path = None
                if write_to_disk:
                    dir_path = os.path.join(
                        write_location, host, dopant, facet, "substrate"
                    )
                    os.makedirs(dir_path, exist_ok=dirs_exist_ok)
                    traj_file_path = os.path.join(dir_path, "input.traj")
                    doped_structure.write(traj_file_path)
                    print(
                        f"{dopant}/{host}({facet}) structure written to {traj_file_path}"
                    )

                saa_structures[host][dopant][facet] = {
                    "structure": doped_structure,
                    "traj_file_path": traj_file_path,
                }
    return saa_structures


def substitute_single_atom_on_surface(
    host_structure: Atoms,
    dopant_element: str,
    place_dopant_at_center: bool = True,
    dopant_magnetic_moment: float = 0.0,
) -> Atoms:
    """
    For a given host (**elemental surface**) structure and a dopant element,
    returns a slab with one host atom on the surface substituted with the
    specified dopant element with a specified magnetic moment.
    Note that for the current implementation (single-atom alloys), there will
    exist only one symmetrically unique site to substitute on the surface of the
    elemental slab.

    Parameters
    ----------

    host_structure (REQUIRED):
        ase.Atoms object of the host slab to be doped.

    dopant_element (REQUIRED):
        String of the elemental species to be substitutionally doped into the
        host structure.

    place_dopant_at_center:
        Boolean specifying whether the single-atom dopant should be placed at
        the center of the unit cell. If False, the dopant atom will be placed at
        the origin.
        Defaults to True.

    dopant_magnetic_moment:
        Float with the initial magnetic moment on the doped single-atom.
        Defaults to no spin polarization (i.e., magnetic moment of 0).

    Returns
    -------

    The elemental slab with a single-atom dopant on the surface as an
    `ase.Atoms` object.

    Raises
    ------

    NotImplementedError
        If multiple symmetrically equivalent sites are found on the surface to dope.
        Note that is intended more as a "guardrail" on current functionality to
        match the maturity/implementation of other modules in `autocat` than an
        actual error. The error should no longer be present when the
        substitution functionality is folded into a more general form.

    """
    tags = host_structure.get_tags()
    constraints = host_structure.constraints
    host_magmoms = host_structure.get_initial_magnetic_moments()

    # convert ase substrate to pymatgen structure
    converter = AseAtomsAdaptor()
    pmg_structure = converter.get_structure(host_structure)

    # find all symmetrically unique site to substitute on
    finder = AdsorbateSiteFinder(pmg_structure)

    # collect all substitution structures and convert them back into ase.Atoms
    pmg_substituted_structures = finder.generate_substitution_structures(dopant_element)
    if len(pmg_substituted_structures) > 1:
        msg = "Multiple symmetrically unique sites to dope found."
        raise NotImplementedError(msg)

    ase_substituted_structure = converter.get_atoms(pmg_substituted_structures[0])
    ase_substituted_structure.set_tags(tags)
    # ensure pbc in xy only
    ase_substituted_structure.pbc = (1, 1, 0)
    # propagate constraints and host magnetization
    ase_substituted_structure.constraints = constraints
    ase_substituted_structure.set_initial_magnetic_moments(host_magmoms)
    # set initial magnetic moment for the dopant atom
    dopant_idx = _find_dopant_index(ase_substituted_structure, dopant_element)
    ase_substituted_structure[dopant_idx].magmom = dopant_magnetic_moment
    # center the single-atom dopant
    if place_dopant_at_center:
        cent_x = (
            ase_substituted_structure.cell[0][0] / 2
            + ase_substituted_structure.cell[1][0] / 2
        )
        cent_y = (
            ase_substituted_structure.cell[0][1] / 2
            + ase_substituted_structure.cell[1][1] / 2
        )
        cent = (cent_x, cent_y, 0)
        ase_substituted_structure.translate(cent)
        ase_substituted_structure.wrap()

    return ase_substituted_structure



import os
import numpy as np
from typing import List
from typing import Tuple
from typing import Dict
from typing import Union
from typing import Any
from typing import Sequence

from ase import Atoms
from ase.build import add_adsorbate
from ase.build import molecule as build_molecule
from ase.data import chemical_symbols
from ase.data import atomic_numbers
from ase.data import covalent_radii
from ase.collections import g2
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.adsorption import AdsorbateSiteFinder
from pymatgen.analysis.local_env import get_neighbors_of_site_with_index

from autocat.data.intermediates import NRR_MOLS
from autocat.data.intermediates import ORR_MOLS


# custom types for readability
RotationOperation = Sequence[Union[float, str]]
RotationOperations = Sequence[RotationOperation]
AdsorptionSite = Dict[str, Sequence[Sequence[float]]]


class AutocatAdsorptionGenerationError(Exception):
    pass


def generate_molecule(
    molecule_name: str = None,
    rotations: RotationOperations = None,
    cell: Sequence[float] = None,
    write_to_disk: bool = False,
    write_location: str = ".",
    dirs_exist_ok: bool = False,
) -> Dict[str, Dict[str, Any]]:
    """
    Generates an `ase.Atoms` object of an isolated molecule.
    If specified, can write out a .traj file containing the isolated molecule
    in a box.

    Parameters
    ----------

    molecule_name (REQUIRED):
        String of the name of the molecule to be generated. Will search in
        `autocat.intermediates` data first and then `ase` g2 database.

    rotations:
        List of rotation operations to be applied to the adsorbate
        molecule/intermediate before being placed on the host surface.

        Example:
        Rotating 90 degrees around the z axis followed by 45 degrees
        around the y-axis can be specified as
            [(90.0, "z"), (45.0, "y")]

        Defaults to [(0, "x")] (i.e., no rotations applied).

    cell:
        List of float specifying the dimensions of the box to place the molecule
        in, in Angstrom.
        Defaults to [15, 15, 15].

    write_to_disk:
        Boolean specifying whether the bulk structures generated should be
        written to disk.
        Defaults to False.

    write_location:
        String with the location where the molecule structure must be written.
        The molecule is written to disk at
        [write_location]/references/[molecule_name]/input.traj

    dirs_exist_ok:
        Boolean specifying whether existing directories/files should be
        overwritten or not. This is passed on to the `os.makedirs` builtin.
        Defaults to False (raises an error if directories corresponding the
        species and crystal structure already exist).

    Returns
    -------

    Dictionary containing Atoms object of the generated molecule and path to
    .traj file if written to disk.

    Example:
    {
        "NH": {"structure": NH_ase_obj, "traj_file_path": "/path/to/NH/traj/file"},
    }
    """
    if molecule_name is None:
        msg = "Molecule name must be specified"
        raise AutocatAdsorptionGenerationError(msg)

    if rotations is None:
        rotations = [[0.0, "x"]]

    if cell is None:
        cell = [15, 15, 15]

    m = None
    if molecule_name in NRR_MOLS:
        m = NRR_MOLS[molecule_name].copy()
    elif molecule_name in ORR_MOLS:
        m = ORR_MOLS[molecule_name].copy()
    elif molecule_name in chemical_symbols:  # atom-in-a-box
        m = Atoms(molecule_name)
    elif molecule_name in g2.names:
        m = build_molecule(molecule_name)

    if m is None:
        msg = f"Unable to construct molecule {molecule_name}"
        raise NotImplementedError(msg)

    for r in rotations:
        m.rotate(r[0], r[1])

    m.cell = cell

    m.center()

    traj_file_path = None
    if write_to_disk:
        dir_path = os.path.join(write_location, "references", f"{molecule_name}")
        os.makedirs(dir_path, exist_ok=dirs_exist_ok)
        traj_file_path = os.path.join(dir_path, "input.traj")
        m.write(traj_file_path)
        print(f"{molecule_name} molecule structure written to {traj_file_path}")

    return {"structure": m, "traj_file_path": traj_file_path}


def generate_adsorbed_structures(
    surface: Union[str, Atoms] = None,
    adsorbates: Union[Dict[str, Union[str, Atoms]], Sequence[str]] = None,
    adsorption_sites: Union[Dict[str, AdsorptionSite], AdsorptionSite] = None,
    use_all_sites: Union[Dict[str, bool], bool] = None,
    site_types: Union[Dict[str, Sequence[str]], Sequence[str]] = None,
    heights: Union[Dict[str, float], float] = None,
    anchor_atom_indices: Union[Dict[str, int], int] = None,
    rotations: Union[Dict[str, RotationOperations], RotationOperations] = None,
    write_to_disk: bool = False,
    write_location: str = ".",
    dirs_exist_ok: bool = False,
) -> Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]:
    """

    Builds structures for reaction intermediates by placing the adsorbate moeity
    on the input surface at specified positions.

    Parameters
    ----------

    surface (REQUIRED):
        Atoms object or path to a file on disk containing the structure of the
        host surface.
        Note that the format of the file must be readable with `ase.io`.

    adsorbates (REQUIRED):
        Dictionary of adsorbate molecule/intermediate names and corresponding
        `ase.Atoms` object or string to be placed on the host surface.

        Note that the strings that appear as values must be in the list of
        supported molecules in `autocat.data.intermediates` or in the `ase` g2
        database. Predefined data in `autocat.data` will take priority over that
        in `ase`.

        Alternatively, a list of strings can be provided as input.
        Note that each string has to be *unique* and available in
        `autocat.data.intermediates` or the `ase` g2 database.

        Example:
        {
            "NH": "NH",
            "N*": "N",
            "NNH": NNH_atoms_obj,
            ...
        }

        OR

        ["NH", "NNH"]

    rotations:
        Dictionary of the list of rotation operations to be applied to each
        adsorbate molecule/intermediate before being placed on the host surface.
        Alternatively, a single list of rotation operations can be provided as
        input to be used for all adsorbates.

        Rotating 90 degrees around the z axis followed by 45 degrees
        around the y-axis can be specified as
            [(90.0, "z"), (45.0, "y")]

        Example:
        {
            "NH": [(90.0, "z"), (45.0, "y")],
            "NNH": ...
        }

        Defaults to [(0, "x")] (i.e., no rotations applied) for each adsorbate
        molecule.

    adsorption_sites:
        Dictionary with labels + list of xy coordinates of sites on the surface
        where each adsorbate must be placed.
        Alternatively, a single dictionary with label and a list of xy
        coordinates can be provided as input to be used for all adsorbates.

        Example:
        {
            "NH": {
                "my_awesome_site_1": [(0.0, 0.0), (0.25, 0.25)],
                "my_awesome_site_2": [(0.0, 1.5)],
                ...
            },
            "NNH": ...
        }

        OR

        {
            "my_default_site": [(0.5, 0.5)],
        }

        Defaults to {"origin": [(0, 0)]} for all adsorbates.

    use_all_sites:
        Dictionary specifying if all symmetrically unique sites on the surface
        must be used for placing each adsorbate. Will generate a number of
        adsorbed structures equal to the number of symmetrically unique sites
        identified using the `AdsorbateSiteFinder` module from `pymatgen`.
        Alternatively, a single Boolean value can be provided as input to be
        used for all adsorbates.

        NB: If True, overrides any sites defined in `adsorption_sites` for each
        adsorbate.

        Defaults to False for all adsorbates.

    site_types:
        Dictionary of the types of adsorption sites to be searched for for each
        adsorbate. Options are "ontop", "bridge", and "hollow".
        Alternatively, a single list of adsorption sites can be provided as
        input to be used for all adsorbates.
        Ignored if `use_all_sites` is False.

        Defaults to ["ontop", "bridge", "hollow"] (if `use_all_sites` is True)
        for all adsorbates.

    heights:
        Dictionary of the height above surface where each adsorbate should be
        placed.
        Alternatively, a single float value can be provided as input to be
        used for all adsorbates.
        If None, will estimate initial height based on covalent radii of the
        nearest neighbor atoms for each adsorbate.

    anchor_atom_indices:
        Dictionary of the integer index of the atom in each adsorbate molecule
        that should be used as anchor when placing it on the surface.
        Alternatively, a single integer index can be provided as input to be
        used for all adsorbates.
        Defaults to the atom at index 0 for each adsorbate molecule.

    write_to_disk:
        Boolean specifying whether the bulk structures generated should be
        written to disk.
        Defaults to False.

    write_location:
        String with the location where the per-species/per-crystal structure
        directories must be constructed and structure files written to disk.
        In the specified write_location, the following directory structure
        will be created:

        adsorbates/[adsorbate_1]/[site_label_1]/[xy_site_coord_1]/input.traj
        adsorbates/[adsorbate_1]/[site_label_1]/[xy_site_coord_2]/input.traj
        ...
        adsorbates/[adsorbate_1]/[site_label_2]/[xy_site_coord_1]/input.traj
        ...
        adsorbates/[adsorbate_2]/[site_label_1]/[xy_site_coord_1]/input.traj
        ...

    dirs_exist_ok:
        Boolean specifying whether existing directories/files should be
        overwritten or not. This is passed on to the `os.makedirs` builtin.
        Defaults to False (raises an error if directories corresponding the
        species and crystal structure already exist).

    Returns
    -------

    ads_structures:
        Dictionary containing all of the reaction structures (with the input
        molecule/intermediate adsorbed onto the surface) as `ase.Atoms` object
        for all adsorbates and adsorption sites.

        Example:

        {
            "NH": {
                "ontop": {
                    "5.345_2.342": {
                        "structure": NH_on_surface_obj,
                        "traj_file_path": "/path/to/NH/adsorbed/on/surface/traj/file"
                    },
                    "1.478_1.230": {
                        ...
                    },
                },
                "hollow": {
                    ...
                },
                ...
            "NNH": {
                ...
            }
        }

    """
    if surface is None:
        msg = "Surface structure must be specified"
        raise AutocatAdsorptionGenerationError(msg)
    elif not isinstance(surface, Atoms):
        msg = f"Unrecognized input type for surface ({type(surface)})"
        raise AutocatAdsorptionGenerationError(msg)

    # Input wrangling for the different types of parameter values allowed for
    # the same function arguments.

    if adsorbates is None or not adsorbates:
        msg = "Adsorbate molecules/intermediates must be specified"
        raise AutocatAdsorptionGenerationError(msg)
    elif isinstance(adsorbates, (list, tuple)):
        adsorbates = {ads_key: ads_key for ads_key in adsorbates}
    elif not isinstance(adsorbates, dict):
        msg = f"Unrecognized input type for adsorbates ({type(adsorbates)})"
        raise AutocatAdsorptionGenerationError(msg)

    if rotations is None:
        rotations = {}
    elif isinstance(rotations, (list, tuple)):
        rotations = {ads_key: rotations for ads_key in adsorbates}
    elif not isinstance(rotations, dict):
        msg = f"Unrecognized input type for rotations ({type(rotations)})"
        raise AutocatAdsorptionGenerationError(msg)

    if adsorption_sites is None:
        adsorption_sites = {}
    elif isinstance(adsorption_sites, dict):
        # check if the input is a single site for all adsorbates vs separate
        # sites for each adsorbate
        if all([ads_key not in adsorption_sites for ads_key in adsorbates]):
            adsorption_sites = {ads_key: adsorption_sites for ads_key in adsorbates}
    elif not isinstance(adsorption_sites, dict):
        msg = f"Unrecognized input type for adsorption_sites ({type(adsorption_sites)})"
        raise AutocatAdsorptionGenerationError(msg)

    if use_all_sites is None:
        use_all_sites = {}
    elif isinstance(use_all_sites, bool):
        use_all_sites = {ads_key: use_all_sites for ads_key in adsorbates}
    elif not isinstance(use_all_sites, dict):
        msg = f"Unrecognized input type for use_all_sites ({type(use_all_sites)})"
        raise AutocatAdsorptionGenerationError(msg)

    if site_types is None:
        site_types = {}
    elif isinstance(site_types, (list, tuple)):
        site_types = {ads_key: site_types for ads_key in adsorbates}
    elif not isinstance(site_types, dict):
        msg = f"Unrecognized input type for site_types ({type(site_types)})"
        raise AutocatAdsorptionGenerationError(msg)

    if heights is None:
        heights = {}
    elif isinstance(heights, float):
        heights = {ads_key: heights for ads_key in adsorbates}
    elif not isinstance(heights, dict):
        msg = f"Unrecognized input type for heights ({type(heights)})"
        raise AutocatAdsorptionGenerationError(msg)

    if anchor_atom_indices is None:
        anchor_atom_indices = {}
    elif isinstance(anchor_atom_indices, int):
        anchor_atom_indices = {ads_key: anchor_atom_indices for ads_key in adsorbates}
    elif not isinstance(anchor_atom_indices, dict):
        msg = f"Unrecognized input type for anchor_atom_indices ({type(anchor_atom_indices)})"
        raise AutocatAdsorptionGenerationError(msg)

    ads_structures = {}
    for ads_key in adsorbates:
        ads_structures[ads_key] = {}
        # generate the molecule if not already input
        adsorbate = adsorbates.get(ads_key)
        if isinstance(adsorbate, str):
            adsorbate = generate_molecule(molecule_name=adsorbate)["structure"]
        # get all adsorption sites for the adsorbate
        ads_adsorption_sites = adsorption_sites.get(ads_key, {"origin": [(0, 0)]})
        ads_use_all_sites = use_all_sites.get(ads_key, False)
        if ads_use_all_sites:
            ads_site_types = site_types.get(ads_key, ["ontop", "bridge", "hollow"])
            ads_adsorption_sites = get_adsorption_sites(
                surface, site_types=ads_site_types
            )
        # for each adsorption site type, for each (xy) coordinate, generate the
        # adsorbated (surface + adsorbate) structure
        for site in ads_adsorption_sites:
            ads_structures[ads_key][site] = {}
            for coords in ads_adsorption_sites[site]:
                # use only xy coords and ignore z if given here (handled by ads_height)
                coords = coords[:2]
                rcoords = np.around(coords, 3)
                scoords = f"{str(rcoords[0])}_{str(rcoords[1])}"
                ads_height = heights.get(ads_key, None)
                ads_anchor_atom_index = anchor_atom_indices.get(ads_key, 0)
                ads_rotations = rotations.get(ads_key, [(0, "x")])
                adsorbed_structure = place_adsorbate(
                    surface=surface,
                    adsorbate=adsorbate,
                    adsorption_site=coords,
                    anchor_atom_index=ads_anchor_atom_index,
                    height=ads_height,
                    rotations=ads_rotations,
                )

                traj_file_path = None
                if write_to_disk:
                    dir_path = os.path.join(
                        write_location, "adsorbates", ads_key, site, scoords
                    )
                    os.makedirs(dir_path, exist_ok=dirs_exist_ok)
                    traj_file_path = os.path.join(dir_path, "input.traj")
                    adsorbed_structure.write(traj_file_path)
                    print(
                        f"Structure with {ads_key} adsorbed at {site}/{scoords}"
                        f" written to {traj_file_path}"
                    )

                ads_structures[ads_key][site].update(
                    {
                        scoords: {
                            "structure": adsorbed_structure,
                            "traj_file_path": traj_file_path,
                        }
                    }
                )

    return ads_structures


def place_adsorbate(
    surface: Atoms = None,
    adsorbate: Atoms = None,
    adsorption_site: Sequence[float] = None,
    rotations: RotationOperations = None,
    anchor_atom_index: int = 0,
    height: float = None,
) -> Atoms:
    """
    Places an adsorbate molecule/intermediate onto a given surface at the specified location.

    Parameters
    ----------

    surface (REQUIRED):
        Atoms object of the host surface.

    adsorbate (REQUIRED):
        Atoms object of the adsorbate molecule/intermediate to be placed on the
        host surface.

    adsorption_site:
        Tuple or list of the xy cartesian coordinates on the surface where the
        adsorbate must be placed.

        Defaults to [0, 0].

    rotations:
        List of rotation operations to be applied to the adsorbate
        molecule/intermediate before being placed on the host surface.

        Example:
        Rotating 90 degrees around the z axis followed by 45 degrees
        around the y-axis can be specified as
            [(90.0, "z"), (45.0, "y")]

        Defaults to [(0, "x")] (i.e., no rotations applied).

    anchor_atom_index:
        Integer index of the atom in the adsorbate molecule that will be used as
        anchor when placing it on the surface.
        Defaults to the atom at index 0.

    height:
        Float specifying the height above surface where adsorbate should be placed.
        If None, will estimate initial height based on covalent radii of the
        nearest neighbor atoms.

    Returns
    -------

    surface:
        Atoms object of the surface with the adsorbate molecule/intermediate
        placed on it as specified.

    """
    if adsorption_site is None:
        adsorption_site = [0, 0]

    if rotations is None:
        rotations = [(0.0, "x")]

    _surface = surface.copy()
    _adsorbate = adsorbate.copy()

    for r in rotations:
        _adsorbate.rotate(r[0], r[1])

    if height is None:
        height = get_adsorbate_height_estimate(
            _surface,
            _adsorbate,
            adsorption_site=adsorption_site,
            anchor_atom_index=anchor_atom_index,
        )

    add_adsorbate(
        _surface,
        _adsorbate,
        height,
        position=adsorption_site,
        mol_index=anchor_atom_index,
    )

    return _surface


def get_adsorbate_height_estimate(
    surface: Atoms = None,
    adsorbate: Atoms = None,
    adsorption_site: Sequence[float] = None,
    anchor_atom_index: int = 0,
    scale: float = 1.0,
) -> float:
    """
    Guess an initial height for the adsorbate to be placed on the surface, by
    summing covalent radii of the nearest neighbor atoms.

    Parameters
    ----------

    surface (REQUIRED):
        `ase.Atoms` object the surface for which adsorbate height must be estimated.

    adsorbate (REQUIRED):
        `ase.Atoms` object of adsorbate to be placed on the surface.

    adsorption_site:
        Tuple or list of the xy cartesian coordinates for where the adsorbate
        would be placed.

        Defaults to [0, 0].

    anchor_atom_index:
        Integer index of the atom in the adsorbate molecule that will be used as
        anchor when placing it on the surface.
        Defaults to the atom at index 0.

    scale:
        Float giving a scale factor to be applied to the calculated bond length.
        For example, scale = 1.1 -> bond length = 1.1 * (covalent_radius_1 + covalent_radius_2)
        Defaults to 1.0 (i.e., no scaling).

    Returns
    -------

    height_estimate:
        Float with the estimated initial height for the adsorbate molecule from
        the surface.
        Returns a default of 1.5 Angstom if the guessing process using
        nearest-neighbor covalent radii fails.
    """
    if adsorption_site is None:
        adsorption_site = (0, 0)

    species, coords = get_adsorbate_slab_nn_list(
        surface=surface, adsorption_site=adsorption_site
    )
    ads_atom_radius = covalent_radii[
        atomic_numbers[adsorbate[anchor_atom_index].symbol]
    ]
    guessed_heights = []
    for sp, coord in zip(species, coords):
        nn_radius = covalent_radii[atomic_numbers[sp]]
        r_dist = scale * (ads_atom_radius + nn_radius)
        position_array = np.array(adsorption_site)
        if (r_dist ** 2 - np.sum((position_array - coord[:2]) ** 2)) >= 0.0:
            height = np.sqrt(r_dist ** 2 - np.sum((position_array - coord[:2]) ** 2))
        else:
            height = np.nan
        guessed_heights.append(height)

    if np.isnan(np.nanmean(guessed_heights)):
        height_estimate = 1.5
    else:
        height_estimate = np.nanmean(guessed_heights)

    return height_estimate


def get_adsorbate_slab_nn_list(
    surface: Atoms = None, adsorption_site: Sequence[float] = None, height: float = 0.5
) -> Tuple[List[str], List[List[float]]]:
    """
    Get list of nearest neighbors for the adsorbate on the surface at the
    specified position.

    Parameters
    ----------

    surface (REQUIRED):
        Atoms object the surface for which nearest neighbors of the adsorbate
        should be identified.

    adsorption_site:
        Tuple or list of the xy cartesian coordinates for where the adsorbate
        would be placed.

        Defaults to [0, 0].

    height:
        Float with the height of the adsorbate molecule from the surface in
        Angstrom.

    Returns
    -------

    species_list, coordinates_list:
        Two lists of length equal to the number of nearest neighbors identified:
        first with the list of species names, and second with the list of
        coordinates.

        Example:
        (["Fe", "Sr"], [[0.4, 0.6], [0.2, 0.9]])
    """
    if adsorption_site is None:
        adsorption_site = (0, 0)

    surf = surface.copy()
    conv = AseAtomsAdaptor()
    add_adsorbate(surf, "X", height=height, position=adsorption_site)
    init_guess = conv.get_structure(surf)
    nn = get_neighbors_of_site_with_index(init_guess, -1)
    species_list = [_nn.species.hill_formula for _nn in nn]
    coord_list = [_nn.coords for _nn in nn]
    return species_list, coord_list


def get_adsorption_sites(
    surface: Atoms = None, site_types: Sequence[str] = None, **kwargs
) -> Dict[str, Sequence[float]]:
    """
    For the given surface, returns all symmetrically unique sites of the
    specified type.

    Uses `pymatgen.analysis.adsorption` module to identify the sites.

    Parameters
    ----------

    surface (REQUIRED):
        Atoms object of the host surface for which the symmetrically unique
        surface sites should be identified.

    site_types:
        List of types of adsorption sites to be searched for. Options are
        "ontop", "bridge", and "hollow".

        Defaults to ["ontop", "bridge", "hollow"].

    **kwargs:
        Other miscellaneous keyword arguments. Will be passed on to
        `AdsorbateSiteFinder.find_adsorption_sites` in `pymatgen`.

    Returns
    -------

    sites:
        Dictionary containing the reference structures with the identified
        sites for each desired site type
    """
    if site_types is None:
        site_types = ["ontop", "bridge", "hollow"]

    converter = AseAtomsAdaptor()
    pmg_structure = converter.get_structure(surface)
    finder = AdsorbateSiteFinder(pmg_structure)
    sites = finder.find_adsorption_sites(positions=site_types, **kwargs)
    # Remove the extraneous "all" returned by default from pymatgen
    sites.pop("all", None)

    return sites



import numpy as np
from ase import Atoms
from ase.io.jsonio import decode as ase_decoder


class Serializable(object):
    """Base abstract class for a serializable object."""

    def to_dict(self):
        """Convert and return object as dictionary."""
        keys = {k.lstrip("_") for k in vars(self)}
        attr = {k: Serializable._to_dict(self.__getattribute__(k)) for k in keys}
        return attr

    @staticmethod
    def _to_dict(obj):
        """Convert obj to a dictionary, and return it."""
        if isinstance(obj, list):
            return [Serializable._to_dict(i) for i in obj]
        elif isinstance(obj, dict):
            return {k: Serializable._to_dict(k) for k in obj}
        elif isinstance(obj, Atoms):
            return ase_decoder(obj)
        elif hasattr(obj, "as_dict"):
            return obj.as_dict()
        else:
            return obj

    @classmethod
    def from_dict(cls, ddict):
        """Construct an object from the input dictionary."""
        return cls(**ddict)



import os
from typing import List
from typing import Dict
from typing import Any

from ase.build import bulk
from ase.data import atomic_numbers
from ase.data import reference_states
from ase.data import ground_state_magnetic_moments
from autocat.data.lattice_parameters import BULK_PBE_FD
from autocat.data.lattice_parameters import BULK_PBE_PW
from autocat.data.lattice_parameters import BULK_BEEFVDW_FD
from autocat.data.lattice_parameters import BULK_BEEFVDW_PW


def generate_bulk_structures(
    species_list: List[str],
    crystal_structures: Dict[str, str] = None,
    default_lat_param_lib: str = None,
    a_dict: Dict[str, float] = None,
    c_dict: Dict[str, float] = None,
    set_magnetic_moments: List[str] = None,
    magnetic_moments: Dict[str, float] = None,
    write_to_disk: bool = False,
    write_location: str = ".",
    dirs_exist_ok: bool = False,
) -> Dict[str, Dict[str, Any]]:
    """
    Generates bulk crystal structures and writes them to separate
    directories, if specified.

    Parameters
    ----------

    species_list (REQUIRED):
        List of chemical symbols of the bulk structures to be constructed.

    cystal_structures:
        Dictionary with crystal structure to be used for each species.
        These will be passed on as input to `ase.build.bulk`. So, must be one
        of sc, fcc, bcc, tetragonal, bct, hcp, rhombohedral, orthorhombic,
        diamond, zincblende, rocksalt, cesiumchloride, fluorite or wurtzite.
        If not specified, the default reference crystal structure for each
        species from `ase.data` will be used.

    default_lat_param_lib:
        String indicating which library the lattice constants should be pulled
        from if not specified in either a_dict or c_dict.
        Defaults to lattice constants defined in `ase.data`.

        Options:
        pbe_fd: parameters calculated using xc=PBE and finite-difference
        beefvdw_fd: parameters calculated using xc=BEEF-vdW and finite-difference
        pbe_pw: parameters calculated using xc=PBE and a plane-wave basis set
        beefvdw_fd: parameters calculated using xc=BEEF-vdW and a plane-wave basis set

        N.B. if there is a species present in species_list that is NOT in the
        reference library specified, it will be pulled from `ase.data`.

    a_dict:
        Dictionary with lattice parameters <a> to be used for each species.
        If not specified, defaults from `default_lat_param_lib` are used.

    c_dict:
        Dictionary with lattice parameters <c> to be used for each species.
        If not specified, defaults from `default_lat_param_lib` are used.

    set_magnetic_moments:
        List of species for which magnetic moments need to be set.
        If not specified, magnetic moments will be set only for Fe, Co, Ni
        (the ferromagnetic elements).

    magnetic_moments:
        Dictionary with the magnetic moments to be set for the chemical
        species listed previously.
        If not specified, default ground state magnetic moments from
        `ase.data` are used.

    write_to_disk:
        Boolean specifying whether the bulk structures generated should be
        written to disk.
        Defaults to False.

    write_location:
        String with the location where the per-species/per-crystal structure
        directories must be constructed and structure files written to disk.

        In the specified write_location, the following directory structure
        will be created:
        [species_1]_bulk_[crystal_structure_1]/input.traj
        [species_1]_bulk_[crystal_structure_2]/input.traj
        ...
        [species_2]_bulk_[crystal_structure_2]/input.traj
        ...
        Defaults to the current working directory.

    dirs_exist_ok:
        Boolean specifying whether existing directories/files should be
        overwritten or not. This is passed on to the `os.makedirs` builtin.
        Defaults to False (raises an error if directories corresponding the
        species and crystal structure already exist).

    Returns
    -------

    Dictionary with bulk structures (as `ase.Atoms` objects) and
    write-location (if any) for each input species.

    Example:

    {
        "Pt": {"structure": Pt_ase_obj, "traj_file_path": "/path/to/Pt/traj/file"},
        "Cr": ...
    }

    """

    lpl = {
        "pbe_fd": BULK_PBE_FD,
        "beefvdw_fd": BULK_BEEFVDW_FD,
        "pbe_pw": BULK_PBE_PW,
        "beefvdw_pw": BULK_BEEFVDW_PW,
    }

    if crystal_structures is None:
        crystal_structures = {}
    if a_dict is None:
        a_dict = {}
    if c_dict is None:
        c_dict = {}
    if set_magnetic_moments is None:
        set_magnetic_moments = ["Fe", "Co", "Ni"]
    if magnetic_moments is None:
        magnetic_moments = {}

    # load crystal structure defaults from `ase.data`, override with user input
    cs_library = {
        species: reference_states[atomic_numbers[species]].get("symmetry")
        for species in species_list
    }
    cs_library.update(crystal_structures)

    # load lattice params <a>, <c> from reference library, override with user input
    a_library = {}
    c_library = {}
    if default_lat_param_lib is not None:
        a_library.update(
            {
                species: lpl[default_lat_param_lib].get(species, {}).get("a")
                for species in species_list
            }
        )
        c_library.update(
            {
                species: lpl[default_lat_param_lib].get(species, {}).get("c")
                for species in species_list
            }
        )
    a_library.update(a_dict)
    c_library.update(c_dict)

    # load magnetic moment defaults from `ase.data`, override with user input
    mm_library = {
        species: ground_state_magnetic_moments[atomic_numbers[species]]
        for species in species_list
    }
    mm_library.update(magnetic_moments)

    bulk_structures = {}
    for species in species_list:
        cs = cs_library.get(species)
        a = a_library.get(species)
        c = c_library.get(species)

        bs = bulk(species, crystalstructure=cs, a=a, c=c)

        if species in set_magnetic_moments:
            bs.set_initial_magnetic_moments([mm_library[species]] * len(bs))

        traj_file_path = None
        if write_to_disk:
            dir_path = os.path.join(write_location, f"{species}_bulk_{cs}")
            os.makedirs(dir_path, exist_ok=dirs_exist_ok)
            traj_file_path = os.path.join(dir_path, "input.traj")
            bs.write(traj_file_path)
            print(f"{species}_bulk_{cs} structure written to {traj_file_path}")

        bulk_structures[species] = {
            "structure": bs,
            "traj_file_path": traj_file_path,
        }

    return bulk_structures



import json
import pkg_resources

__all__ = ["BULK_PBE_FD", "BULK_BEEFVDW_FD", "BULK_PBE_PW", "BULK_BEEFVDW_PW"]
"""
Calculator Settings:

    BULK_PBE_FD, BULK_BEEFVDW_FD:

        Obtained via fits to an equation of state
        (https://wiki.fysik.dtu.dk/ase/ase/eos.html)

        FCC/BCC
        h = 0.16, kpts = (12,12,12)
        fit to an SJ EOS

        HCP
        h=0.16, kpts = (12,12,6)
        fit to a Birch-Murnaghan EO

    BULK_PBE_PW, BULK_BEEFVDW_PW:

        Obtained using the Exponential Cell Filter to minimize the stress tensor and atomic forces
        (https://wiki.fysik.dtu.dk/ase/ase/constraints.html#the-expcellfilter-class)

        FCC/BCC
        mode=PW(550), kpts = (12,12,12), fmax = 0.05 eV/A

        HCP
        mode=PW(550), kpts = (12,12,6), fmax = 0.05 eV/A

"""

raw_bulk_pbe_fd = pkg_resources.resource_filename(
    "autocat.data.lattice_parameters", "bulk_pbe_fd.json"
)

with open(raw_bulk_pbe_fd) as fr:
    BULK_PBE_FD = json.load(fr)


raw_bulk_beefvdw_fd = pkg_resources.resource_filename(
    "autocat.data.lattice_parameters", "bulk_beefvdw_fd.json"
)

with open(raw_bulk_beefvdw_fd) as fr:
    BULK_BEEFVDW_FD = json.load(fr)


raw_bulk_pbe_pw = pkg_resources.resource_filename(
    "autocat.data.lattice_parameters", "bulk_pbe_pw.json"
)

with open(raw_bulk_pbe_pw) as fr:
    BULK_PBE_PW = json.load(fr)


raw_bulk_beefvdw_pw = pkg_resources.resource_filename(
    "autocat.data.lattice_parameters", "bulk_beefvdw_pw.json"
)

with open(raw_bulk_beefvdw_pw) as fr:
    BULK_BEEFVDW_PW = json.load(fr)



import os
from setuptools import setup
from setuptools import find_packages


version_file = os.path.join(os.path.dirname(__file__), "src", "autocat", "VERSION.txt")
with open(version_file, "r") as fr:
    version = fr.read().strip()


setup(
    name="autocat",
    version=version,
    description="Tools for automated generation of catalyst structures and sequential learning",
    url="https://github.com/aced-differentiate/auto_cat",
    author="Lance Kavalsky",
    author_email="lkavalsk@andrew.cmu.edu",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.9",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=["numpy<=1.22.0", "ase", "pymatgen<=2022.0.17", "fire",],
    include_package_data=True,
)
